<?php

namespace Stackify\Util;

/**
 * Provides a static method to call functions synchronously using lockfiles.
 */
class Sync
{
    /**
     * Default lock file
     */
    public static $lock = '.lock';
    private static $locks = array();

    /**
     * Call a function in a mutually exclusive way using a lockfile.
     * A process will only block other processes and never block itself,
     * so you can safely nest synchronized operations.
     * @throws \Exception
     */
    public static function call($func, $lock = null)
    {
        $lock = self::path($lock);
        // just call function if lock already acquired
        if (isset(self::$locks[$lock])) {
            return $func();
        }
        try {
            // acquire lock
            $handle = fopen($lock, 'w');
            if (!$handle || !flock($handle, LOCK_EX)) {
                throw new \RuntimeException('Unable to lock on "' . $lock . '"');
            }
            self::$locks[$lock] = true;
            // call function and release lock
            $return = $func();
            self::release($lock, $handle);
            return $return;
        } catch (\Exception $e) {
            // release lock and rethrow any exception
            self::release($lock, $handle);
            throw $e;
        }
    }

    /**
     * Get absolute path to a lock file
     */
    public static function path($lock = null)
    {
        if (!isset($lock)) {
            $lock = self::$lock;
        }
        if (!is_file($lock)) {
            touch($lock);
            chmod($lock, 0777);
        }
        return realpath($lock);
    }

    /**
     * Set of already aquired locks
     */
    private static function release($lock, $handle)
    {
        unset(self::$locks[$lock]);
        if ($handle) {
            flock($handle, LOCK_UN);
            fclose($handle);
        }
    }
}

namespace Stackify\Profiler;

/**
 * Handles link to PHP Extension.
 * @package Stackify\Profiler
 */
class StackifyExtension
{
    public function getTraceFrames()
    {
        $frames = stackify_get_frames();

        if (!is_array($frames)) {
            return array();
        }

        return $frames;
    }

    public function watchSpan($function_name, $category)
    {
        return stackify_span_watch($function_name, $category);
    }
}

namespace Stackify;

class Util
{
    const RUM_URL_ID_REGEX = '/^(\d+)$/';
    const RUM_URL_GUID_REGEX = '/^(\b[A-F0-9]{8}(?:-[A-F0-9]{4}){3}-[A-F0-9]{12}\b)$/i';
    const RUM_URL_EMAIL_REGEX = '/^((([!#$%&\'*+\-\/=?^_`{|}~\w])|([!#$%&\'*+\-\/=?^_`{|}~\w][!#$%&\'*+\-\/=?^_`{|}~\.\w]{0,}[!#$%&\'*+\-\/=?^_`{|}~\w]))[@]\w+([-.]\w+)*\.\w+([-.]\w+)*)$/';
    const RUM_URL_IP_REGEX = '/^(25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.(25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.(25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.(25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)$/';

    private static $isWindows = null;
    private static $isPrefixEnabled = null;

    /**
     * Mask string based on reporting url rules
     *
     * @param string $rawString Payload string
     *
     * @return string
     */
    public static function maskString($rawString)
    {
        if (preg_match(self::RUM_URL_ID_REGEX, $rawString)) {
            return '{id}';
        }

        if (preg_match(self::RUM_URL_GUID_REGEX, $rawString)) {
            return '{guid}';
        }

        if (preg_match(self::RUM_URL_EMAIL_REGEX, $rawString)) {
            return '{email}';
        }

        if (preg_match(self::RUM_URL_IP_REGEX, $rawString)) {
            return '{ip}';
        }

        $semicolonIndex = strpos($rawString, ';');
        if ($semicolonIndex !== false) {
            return substr($rawString, 0, $semicolonIndex);
        }

        return $rawString;
    }

    /**
     * Mask reporting url
     *
     * @param string $url Payload string
     *
     * @return string
     */
    public static function maskReportingUrl($url)
    {
        if (empty($url)) {
            return '/';
        }

        $parts = array();
        if (!empty($url)) {
            $parts = explode('/', $url);
        }

        if (!count($parts)) {
            return '/';
        }

        $maskedParts = array();
        foreach ($parts as $part) {
            $maskedParts[] = self::maskString($part);
        }

        $maskedUrl = join('/', $maskedParts);
        $maskedUrlLen = strlen($maskedUrl);

        if ($maskedUrlLen === 1) {
            return $maskedUrl;
        }

        if (substr($maskedUrl, -1, 1) === '/') {
            $trimmedUrl = substr($maskedUrl, 0, $maskedUrlLen - 1);

            if (empty($trimmedUrl)) {
                return '/';
            }

            return $trimmedUrl;
        }

        return $maskedUrl;
    }

    /**
     * Check if OS is Windows
     *
     * @return boolean
     */
    public static function isWindows()
    {
        if (static::$isWindows == null) {
            static::$isWindows = strtoupper(substr(PHP_OS, 0, 3)) === 'WIN';
        }
        return static::$isWindows;
    }

    /**
     * Make symlink
     *
     * @param string $target target file/directory
     * @param string $link   new link
     *
     * @return mixed
     */
    public static function makeSymlink($target, $link)
    {
        if (static::$isWindows) {
            $output = array();
            $resultCode = null;
            $command = "mklink /j $link $target";
            if (exec($command, $output, $resultCode) === false) {
                Profiler::log(1, "Util::makeSymlink - Windows - Unable to make link - Target: $target to Link: $link - Output: ". print_r(array($output, $resultCode), true));
            }
            return true;
        }

        return symlink($target, $link);
    }

    /**
     * Check if file or folder is symlink
     *
     * @param string $file file or directory to be removed
     *
     * @return boolean
     */
    public static function isLink($file)
    {
        if (static::$isWindows) {
            return readlink($file) !== false;
        }

        return is_link($file);
    }

    /**
     * Remove file or directory
     *
     * @param string $file file or directory to be removed
     *
     * @return mixed
     */
    public static function unLink($file)
    {
        if (static::$isWindows && static::isLink($file)) {
            return rmdir($file) !== false;
        }
        return unlink($file);
    }

    /**
     * Check if prefix is enabled
     *
     * @return boolean
     */
    public static function isPrefixEnabled()
    {
        if (static::$isPrefixEnabled === null) {
            $prefixEnabled = strval(ini_get("stackify.prefix_enabled"));
            static::$isPrefixEnabled = $prefixEnabled != null && $prefixEnabled === "1";
        }

        return static::$isPrefixEnabled;
    }

    /**
     * Checks if file exists, it's a file, file is readable, and file is not empty
     *
     * @param $file File to be checked
     *
     * @return boolean
     */
    public static function isFileNotEmpty($file)
    {
        if($file !== "" && file_exists($file) && is_file($file) && is_readable($file) && filesize($file) > 0)
        {
            return true;
        } else
            return false;
    }
}

class Profiler
{
    const MODE_DISABLED = 0;
    const MODE_FULL = 2;
    const DEFAULT_RUM_SCRIPT_URL = 'https://stckjs.stackify.com/stckjs.js';
    const DEFAULT_RUM_KEY = '';

    // default profiler configuration
    public static $defaultOptions = array(
        'ignored_functions' => array(
            'call_user_func',
            'call_user_func_array',
            'array_filter',
            'array_map',
            'array_reduce',
            'array_walk',
            'array_walk_recursive',
            'Symfony\Component\DependencyInjection\Container::get',
        ),
        'exception_function' => null,
        'framework' => null,
        'transaction_functions' => array(),
        'transaction_callback' => null,
        'watches' => array()
    );

    public static $trace;
    private static $stackifyExtension;
    private static $error = false;
    private static $mode = self::MODE_DISABLED;
    private static $applicationName;
    private static $environmentName;
    private static $transport;
    private static $transportHttpEndpoint;
    private static $exceptions = array();
    private static $transactionName;

    public static $traceMapByParentFrameId;
    private static $max_files = 10;
    private static $logFilePath =  null;
    private static $logFilePathExist = null;
    private static $errorLogExist = null;
    private static $errorLog = null;
    private static $errorLogFile = null;
    private static $isCli = false;
    private static $logFilePathWritable = null;
    public static $customUrlTransactionName = null;
    public static $hasExecutedTransactionFunction = false;
    private static $rumScriptUrl = self::DEFAULT_RUM_SCRIPT_URL;
    private static $rumKey = self::DEFAULT_RUM_KEY;

    private static $options = null;

    /**
     * Custom error handler to log profiler specific errors to stackify log file - to avoid noise in standard php error logging.
     * Only runs when debug is enabled
     */
    public static function stackifyErrorHandler($errno, $errstr, $errfile, $errline) {
        if (self::isDebug()) {
            self::log(1, $errstr . " : " .  $errfile . " : " . $errno . " : " . $errline);
        }
    }

    /**
     * Check if debug mode is enabled on the profiler
     * Checks ini and environment variables
     */
    public static function isDebug() {
        $debug_enabled = strval(ini_get("stackify.debug_enabled"));
        if ($debug_enabled != NULL && $debug_enabled == "1") {
            return true;
        }
        $debug_enabled = strval(getenv('STACKIFY_DEBUG_ENABLED'));
        if ($debug_enabled != NULL && strtoupper($debug_enabled) == "TRUE") {
            return true;
        }
        return false;
    }

    /**
     * Gets the configured track all
     * Checks ini and environment variables
     */
    public static function isTrackAll() {
        $track_all_enabled = strval(ini_get("stackify.track_all_enabled"));
        if ($track_all_enabled != NULL && $track_all_enabled == "1") {
            return true;
        }
        $track_all_enabled = strval(getenv('STACKIFY_TRACK_ALL_ENABLED'));
        if ($track_all_enabled != NULL && strtoupper($track_all_enabled) == "TRUE") {
            return true;
        }
        return false;
    }

    /**
     * Add custom transaction name for current transaction.
     * @param $transactionName
     */
    public static function setTransactionName($transactionName)
    {
        self::$transactionName = $transactionName;
    }

    /**
     * Add more ignore functions to profiling options.
     *
     * @param array<string> $functionNames
     * @return void
     */
    public static function addIgnoreFunctions(array $functionNames)
    {
        foreach ($functionNames as $functionName) {
            self::$defaultOptions['ignored_functions'][] = $functionName;
        }
    }

    /**
     * Shutdown profiler
     *
     * @return void
     */
    public static function shutdown()
    {
        if (self::$mode === self::MODE_DISABLED) {
            return;
        }

        try {
            $lastError = error_get_last();
            if (self::isReportableError($lastError)) {
                if (version_compare(phpversion(), '8', '>=')) {
                    define('E_DONT_BAIL', 1 << 15);
                } else {
                    define('E_DONT_BAIL', 1 << 0);
                }
                $lastErrorStackify = stackify_error_get_last();
                $lastErrorStackifyTracked = stackify_error_get_last_tracked();
                $addException = false;

                if ($lastErrorStackifyTracked) {
                    if ($lastErrorStackify // We are able to track the last error (so we compare it)
                        && ($lastError['type'] & ~E_DONT_BAIL) == ($lastErrorStackify['type'] & ~E_DONT_BAIL)
                        && $lastError['file'] == $lastErrorStackify['file']
                        && $lastError['line'] == $lastErrorStackify['line']
                    ) {
                        $addException = true; // Exception was tracked and has same details
                    }
                } else {
                    $addException = true; // Exception was not tracked, we report it
                }

                if ($addException) {
                    $lastError['trace'] = function_exists('stackify_fatal_backtrace') ? stackify_fatal_backtrace() : null;
                    array_push(self::$exceptions, $lastError);
                }
            }

            self::stop();
        } catch (\Exception $e) {
            self::log(3, "Stackify\Profiler::shutdown - Exception: ". $e->getMessage());
        }
    }

    /**
     * Returns in passed in error should be reported
     *
     * @param  $error Error object
     *
     * @return bool
     */
    public static function isReportableError($error)
    {
        $stackifyErrorReportingLevel = self::_getErrorReportingLevelFilter();
        if ($error && $error["type"]) {
            if ($stackifyErrorReportingLevel !== null) {
                // check if reporting level contains this error type
               return (($stackifyErrorReportingLevel & $error["type"]) === $error["type"]);
            } else {
                // default types to include if error reporting level is not present
                return self::isFatalError($error["type"]);
            }
        }
        return false;
    }

    /**
     * Check if error type is fatal
     *
     * @param int $type Error type
     *
     * @return boolean
     */
    public static function isFatalError($type)
    {
        switch ($type)
        {
        case E_ERROR:
        case E_CORE_ERROR:
        case E_COMPILE_ERROR:
        case E_USER_ERROR:
        case E_PARSE:
            return true;
            break;
        }
        return false;
    }

    /**
     * Stop all profiling actions and submit collected data.
     */
    public static function stop()
    {
        if (self::$mode === self::MODE_DISABLED) {
            return;
        }

        $mode = self::$mode;

        // stop profiler
        if (($mode & self::MODE_FULL) > 0 || self::$error) {
            stackify_disable();
        }
        self::$mode = self::MODE_DISABLED;

        if (!self::$hasExecutedTransactionFunction) {
            self::sendTrace();
        }
     }

    /**
     * Adds AWS Lambda properties to traces if available
     */
     public static function applyAWSLambdaProperties()
     {
        $lambdaFunctionName = getenv('AWS_LAMBDA_FUNCTION_NAME');
        if (!empty($lambdaFunctionName)) {
            $lambdaHandler = getenv('_HANDLER');
            self::$trace['call'] = $lambdaFunctionName . "::" . $lambdaHandler;
            self::$trace['props']['REPORTING_URL'] = $lambdaFunctionName;
            self::$trace['props']['APPLICATION_FILESYSTEM_PATH'] = "/";
        }
     }

    /**
     * Converts frames to trace then writes to transport
     */
    public static function sendTrace()
    {
        // get frames
        $frames = self::$stackifyExtension->getTraceFrames();

        if (!is_array($frames)) {
            return;
        }

        // link frames and convert to trace
        self::$trace = self::linkFrames($frames);
        // apply aws lambda properties if available
        self::applyAWSLambdaProperties();

        $exceptions_count = sizeof(self::$exceptions);
        if (sizeof(self::$exceptions) > 0) {
            self::$trace['exceptions'] = array();
            for ($i = 0; $i < $exceptions_count;$i++) {

                $trace_exception = array();
                $exception_message_parts = array();
                if (!empty(self::$exceptions[$i]['message'])) {
                    $exception_message_parts = explode("\n", self::$exceptions[$i]['message']);
                }

                if (sizeof($exception_message_parts) >= 1) {

                    $trace_exception['Timestamp'] = strval((int) self::$trace['reqEnd']);
                    $trace_exception['Exception'] = $exception_message_parts[0];
                    $trace_exception['Message'] = $exception_message_parts[0];

                    if (sizeof($exception_message_parts) >= 3) {
                        $trace_exception['CaughtBy'] = $exception_message_parts[2];
                    } else {
                        $trace_exception['CaughtBy'] = "Unknown";
                    }

                    $trace_exception['Frames'] = array();
                    foreach ($exception_message_parts as $exception_message_part_key => $exception_message_part) {
                        $exceptionFrame = array();
                        $exceptionFrame['Method'] = $exception_message_part;
                        array_push($trace_exception['Frames'], $exceptionFrame);
                    }

                    array_push(self::$trace['exceptions'], $trace_exception);
                }
            }
        }

       $output_json = self::safe_json_encode(self::$trace, JSON_UNESCAPED_UNICODE | JSON_UNESCAPED_SLASHES | JSON_NUMERIC_CHECK);

       if ($output_json === false ) {
            try {
                $serializedString = serialize(self::$trace);
                $errorMaxLength = ini_get("log_errors_max_len");

                if ($errorMaxLength !== false && strlen($serializedString) < $errorMaxLength) {
                    self::log(3, "Stackify\Profiler::sendTrace - json_encode('trace') - Payload: ".$errorMaxLength);
                } else {
                    self::log(3, "Stackify\Profiler::sendTrace - json_encode('trace') - Payload Error: Greater than the error text limit (log_errors_max_len)");
                }
            } catch (\Exception $e) {
                self::log(3, "Stackify\Profiler::sendTrace - json_encode('trace') - serialize('trace') -> Exception: ". $e->getMessage());
            }

        } else {

            if (self::$transport == null || self::$transport == "log") {
                try {
                    // write to trace log
                    $prefix = gmdate("Y-m-d, H:i:s.u") . "> ";
                    $output_final = $prefix . $output_json . "\n";
                    self::writeToTraceLog($output_final);
                } catch (\Exception $e) {
                    self::log(3, "Stackify\Profiler::sendTrace - error writing trace to log file.". $e->getMessage());
                }
            } else if (self::$transport == "console") {
                try {
                    // write to console
                    $output_json_gzip = gzencode($output_json, 6);
                    $output_json_gzip_base64 = base64_encode($output_json_gzip);
                    $prefix = "STACKIFY-TRACE: ";
                    $output_final = $prefix . $output_json_gzip_base64 . "\n";
                    echo($output_final);
                } catch (\Exception $e) {
                    self::log(3, "Stackify\Profiler::sendTrace - error writing trace console.". $e->getMessage());
                }
            } else if (self::$transport == "agent_http") {
                try {
                    if (extension_loaded('curl')) {
                        $output_json = "[" . $output_json . "]";
                        $url = self::$transportHttpEndpoint . "/traces";
                        $ch = curl_init();
                        curl_setopt($ch, CURLOPT_CUSTOMREQUEST, "POST");
                        curl_setopt($ch, CURLOPT_URL, $url);
                        curl_setopt($ch, CURLOPT_SSL_VERIFYHOST, 0);
                        curl_setopt($ch, CURLOPT_SSL_VERIFYPEER, 0);
                        curl_setopt($ch, CURLOPT_POSTFIELDS, $output_json);
                        curl_setopt($ch, CURLOPT_RETURNTRANSFER, true);
                        curl_setopt($ch, CURLOPT_HTTPHEADER, array(
                                'Content-Type: application/json',
                                'Content-Length: ' . strlen($output_json))
                        );
                        $result = curl_exec($ch);
                        curl_close($ch);
                    } else {
                        self::log(3, "Stackify\Profiler::sendTrace - curl extension is required for the agent_http transport");
                    }
                } catch (\Exception $e) {
                    self::log(3, "Stackify\Profiler::sendTrace - error posting trace to [" . self::$transportHttpEndpoint . "].". $e->getMessage());
                }
            }
        }

        self::$traceMapByParentFrameId = null;
        self::$customUrlTransactionName = null;
        self::$trace = null; // free memory
        unset($frames); // Dereference avoid cow violation
    }

    /**
     * Fixes json_encode posiblle problems
     * http://php.net/manual/en/function.json-last-error.php#121233
     * https://stackoverflow.com/questions/10199017/how-to-solve-json-error-utf8-error-in-php-json-decode
     */
    public static function safe_json_encode($value, $options = 0, $depth = 512, $utfErrorFlag = false) {
        if ((defined('PHP_MAJOR_VERSION') && PHP_MAJOR_VERSION >= 5) && (defined('PHP_MINOR_VERSION') && PHP_MINOR_VERSION >= 6) ) {
            $encoded = json_encode($value, $options, $depth);
        } else {
            $encoded = json_encode($value, $options);
        }

        switch (json_last_error()) {
            case JSON_ERROR_NONE:
                return $encoded;
            case JSON_ERROR_DEPTH:
                self::log(3, "Stackify\Profiler::stop - json_encode('trace') - Maximum stack depth exceeded!");
                self::log(3, "Stackify\Profiler::stop - json_encode('trace') - ErrorMessage: ".json_last_error_msg());
                return false;
            case JSON_ERROR_STATE_MISMATCH:
                self::log(3, "Stackify\Profiler::stop - json_encode('trace') - Underflow or the modes mismatch!");
                self::log(3, "Stackify\Profiler::stop - json_encode('trace') - ErrorMessage: ".json_last_error_msg());
                return false;
            case JSON_ERROR_CTRL_CHAR:
                self::log(3, "Stackify\Profiler::stop - json_encode('trace') - Unexpected control character found!");
                self::log(3, "Stackify\Profiler::stop - json_encode('trace') - ErrorMessage: ".json_last_error_msg());
                return false;
            case JSON_ERROR_SYNTAX:
                self::log(3, "Stackify\Profiler::stop - json_encode('trace') - Syntax error, malformed JSON!");
                self::log(3, "Stackify\Profiler::stop - json_encode('trace') - ErrorMessage: ".json_last_error_msg());
                return false;
            case JSON_ERROR_UTF8:
                $clean = self::utf8ize($value);
                if ($utfErrorFlag) {
                    self::log(3, "Stackify\Profiler::stop - json_encode('trace') - UTF8 encoding error!");
                    self::log(3, "Stackify\Profiler::stop - json_encode('trace') - ErrorMessage: ".json_last_error_msg());
                    return false;
                }
                return self::safe_json_encode($clean, $options, $depth, true);
            default:
                self::log(3, "Stackify\Profiler::stop - json_encode('trace') - Unknown error!");
                return false;
        }

    }

    /**
     * Fixes json_encode JSON_ERROR_UTF8 error
     * http://php.net/manual/en/function.json-last-error.php#115980
     */
    public static function utf8ize($mixed) {
        if (is_array($mixed)) {
            foreach ($mixed as $key => $value) {
                $mixed[$key] = self::utf8ize($value);
            }
        } else if (is_string ($mixed)) {
            return utf8_encode($mixed);
        }
        return $mixed;
    }

    /**
     * Handles any post-processing on frames
     */
    public static function post_process_frames($frames)
    {
        if (!is_array($frames)) {
            return;
        }

        $processed_frames = array();

        foreach ($frames as $frame)  {

            if (isset($frame['props']['PREFIX_RESPONSE_BODY'])) {
                $response_headers = array();
                $response_body = $frame['props']['PREFIX_RESPONSE_BODY'];
                $response_body_parts = array();
                if (!empty($response_body)) {
                    $response_body_parts = explode( "\r\n", $response_body);
                }

                foreach ($response_body_parts as $response_body_part) {
                    if (empty($response_body_part)) {
                        break;
                    } else {
                        $del_position = strpos($response_body_part, ":");
                        if($del_position !== false) {

                            $key = substr($response_body_part, 0, $del_position);
                            $value = substr($response_body_part, $del_position + 1, strlen($response_body_part));

                            if (!empty($key) && !empty($value)) {
                                $response_header = array();
                                $response_header[trim($key)] = trim($value);
                                array_push($response_headers, $response_header);
                            }
                        }
                    }
                }

                $frame['props']['PREFIX_RESPONSE_HEADERS'] = self::safe_json_encode($response_headers);
            }

            array_push($processed_frames, $frame);
        }

        return $processed_frames;
    }

    /**
     * Removes slow frames
     */
    public static function trim_frames(&$frames)
    {
      if (!is_array($frames)) {
          return;
      }

      // [frame_id] -> [frame]  map
      $frame_id_map = array();
      foreach ($frames as $frame_key => $frame) {
        $frame_id_map[$frame['props']['FRAME_ID']] = $frame;
      }

      // [frame_id] -> [parent_frame_id] map
      $frame_id_to_parent_frame_id = array();
      foreach ($frames as $frame_key => $frame) {
        if (isset($frame['props']['PARENT_FRAME_ID'])) {
          $parent_frame_id = $frame['props']['PARENT_FRAME_ID'];
          $parent_frame = $frame_id_map[$parent_frame_id];
          $parent_parent_frame_id = 0;
          if (isset($parent_frame['props']['PARENT_FRAME_ID'])) {
            $parent_parent_frame_id = $parent_frame['props']['PARENT_FRAME_ID'];
          }
          $frame_id_to_parent_frame_id[$frame['props']['FRAME_ID']] = $parent_parent_frame_id;
        }
      }

      // remove frames if timings are missing
      foreach ($frames as $frame_key => $frame) {
        if (!array_key_exists('reqBegin', $frame) || !array_key_exists('reqBegin', $frame)) {
          unset($frames[$frame_key]);
        }
      }

      // remove fast frames
      foreach ($frames as $frame_key => $frame) {
        $frame_id = intval($frame['props']['FRAME_ID']);
        if ($frame_id > 0) {  // skip root frame
          $req_start_ms = doubleval($frame['reqBegin']);
          // edge case where reqEnd is not set - avoid error by setting to reqBegin (likely will be removed as a fast frame)
          $req_end_ms = array_key_exists('reqEnd', $frame) ? doubleval($frame['reqEnd']) : 0;
          $elapsed_ms = $req_end_ms - $req_start_ms;
          $property_count = sizeof($frame['props']);
          if ($elapsed_ms < 1 && $property_count < 3) {
            unset($frames[$frame_key]);
          }
        }
      }

      // trimmed [frame_id] -> [frame]  map
      $trimmed_trace_map = array();
      foreach ($frames as $frame_key => $frame) {
        $trimmed_trace_map[$frame['props']['FRAME_ID']] = $frame;
      }

      // update [parent_frame_id] on needed frames
      foreach ($frames as $frame_key => $frame) {
        if (isset($frame['props']['PARENT_FRAME_ID'])) {
          $parent_frame_id = $frame['props']['PARENT_FRAME_ID'];
          if (isset($parent_frame_id)) {
            $count = 0;
            while (!isset($trimmed_trace_map[$parent_frame_id]) && $count < 25) {
              $count = $count + 1; // limit look loop
              $parent_frame_id = $frame_id_to_parent_frame_id[$parent_frame_id];
            }
            if (!isset($parent_frame_id)) {
              $parent_frame_id = 0;
            }
            $frame['props']['PARENT_FRAME_ID'] = $parent_frame_id;
          }
        }
      }
    }

    /**
     * Converts array of frames into a trace tree.
     * @param $frames
     * @return trace
     */
    public static function linkFrames($frames)
    {
        $show_fast_functions = false;
        if (isset(self::$defaultOptions['show_fast_functions']) && self::$defaultOptions['show_fast_functions'] == 1) {
            $show_fast_functions = true;
        }

        $trace = $frames[0];

        if (!$show_fast_functions) {
          self::trim_frames($frames);
        }

        $frames = self::post_process_frames($frames);

        self::setupFrameStack($trace, $frames);

        self::setupTraceProperties($trace);
        return $trace;
    }

    /**
     * Add trace level properties.
     */
    static function setupTraceProperties(&$trace)
    {
        if (!is_array($trace)) {
            return;
        }

        $trace['props']['APPLICATION_PATH'] = "/";
        $trace['props']['APPLICATION_FILESYSTEM_PATH'] = static::getServerValue('DOCUMENT_ROOT');
        $trace['props']['APPLICATION_NAME'] = self::$applicationName;
        $trace['props']['APPLICATION_ENV'] = self::$environmentName;
        $trace['props']['CATEGORY'] = 'PHP';
        $trace['props']['THREADID'] = 0;

        if (self::$isCli) {
            $trace['call'] = 'CLI';
            if (empty($trace['props']['APPLICATION_FILESYSTEM_PATH'])) {
                $trace['props']['APPLICATION_FILESYSTEM_PATH'] = __DIR__;
            }
        }

        if (!self::$isCli) {
            $trace['props']['URL'] = self::full_url();
        }

        $trace['props']['REPORTING_URL'] = self::getReportingUrl();

        $isRum = false;
        if (isset($trace['props']['ISRUM']) && $trace['props']['ISRUM']) {
            $isRum = $trace['props']['ISRUM'];
        }

        $trace['props']['ISRUM'] = $isRum ? "TRUE" : "FALSE";

        if (Util::isPrefixEnabled()) {
            $trace['props']['PREFIX'] = "TRUE";

            // add request headers
            $trace['props']['PREFIX_REQUEST_HEADERS'] =  self::safe_json_encode(static::getRequestHeaders(), JSON_UNESCAPED_UNICODE | JSON_UNESCAPED_SLASHES | JSON_NUMERIC_CHECK);

            // add request body
            $request_body = file_get_contents('php://input');
            $trace['props']['PREFIX_REQUEST_BODY'] = $request_body;
            $trace['props']['PREFIX_REQUEST_SIZE_BYTES'] = strlen($request_body);

            // add response headers
            $response_headers = array();
            foreach (headers_list() as $header) {
                $header_parts = array();
                if (!empty($header)) {
                    $header_parts = explode(':', $header);
                }

                if (count($header_parts) == 2) {
                    $name = $header_parts[0];
                    $value = $header_parts[1];
                    $response_header = [
                        $name => $value
                    ];
                    array_push($response_headers, $response_header);
                }
            }
            $trace['props']['PREFIX_RESPONSE_HEADERS'] = self::safe_json_encode($response_headers, JSON_UNESCAPED_UNICODE | JSON_UNESCAPED_SLASHES | JSON_NUMERIC_CHECK);
        }

        if (!self::$isCli) {
            $trace['props']['METHOD'] = static::getServerValue('REQUEST_METHOD');
            $trace['props']['STATUS'] = http_response_code();
        }

        $trace['props']['TRACETYPE'] = self::$isCli ? 'TASK' : 'WEBAPP';
        $trace['props']['TRACE_SOURCE'] = 'PHP';
        $trace['props']['TRACE_TARGET'] = 'RETRACE';
        $trace['props']['HOST_NAME'] = gethostname();

        $trace['props']['OS_TYPE'] = Util::isWindows() ? 'WINDOWS': 'LINUX';

        // handle if getmypid is disabled - default to 0
        $pid = getmypid();
        $pid = isset($pid) ? $pid : '0';

        $trace['props']['PROCESS_ID'] = $pid;
        $trace['props']['TRACE_VERSION'] = "2.0";

    }

    /**
     * Returns customized reporting url based from stackify.json settings.
     * @param $reporting_url
     */
    public static function custom_url()
    {
        if (self::$customUrlTransactionName != null) {
            return self::$customUrlTransactionName;
        }
    }
    /**
     * Returns the 1st element of the path passed in (will use back or forward slash).
     */
    static function get_first_path_element($path)
    {
        if (strpos($path, '/') !== false) {
            $application_path_elements = explode('/', $path);
            $application_path_elements = array_filter($application_path_elements);
            return array_values($application_path_elements)[0];
        } else if (strpos($path, '\\') !== false) {
            $application_path_elements = explode('\\', $path);
            $application_path_elements = array_filter($application_path_elements);
            return array_values($application_path_elements)[0];
        }

        return $path;
    }

    /**
     * Returns full url for current application request.
     */
    static function full_url()
    {
        return self::remove_trailing_slash(self::url_origin() . static::getServerValue('REQUEST_URI'));
    }

    /**
     * Utility to remove trailing forward or backslash from value passed in.
     */
    static function remove_trailing_slash($v)
    {
        if (substr($v, -1) == '/') {
            $v = substr($v, 0, -1);
        } else if (substr($v, -1) == '\\') {
            $v = substr($v, 0, -1);
        }
        return $v;
    }

    /**
     * Returns origin portion of current application request.
     */
    static function url_origin()
    {
        $use_forwarded_host = static::useForwardedHost();
        $forwardedHost = static::getServerValue('HTTP_X_FORWARDED_HOST');
        $https = static::getServerValue('HTTPS');
        $host = static::getServerValue('HTTP_HOST');
        $serverProtocol = static::getServerValue('SERVER_PROTOCOL');

        $ssl = (!empty($https) && $https == 'on');
        $protocol = '';

        if (!empty($serverProtocol)) {
            $sp = strtolower($serverProtocol);
            $protocol = substr($sp, 0, strpos($sp, '/')) . (($ssl) ? 's' : '');
        }

        $port = static::getServerValue('SERVER_PORT');
        $port = ((!$ssl && $port == '80') || ($ssl && $port == '443')) ? '' : ':' . $port;
        $hostname = $use_forwarded_host && $forwardedHost ? $forwardedHost : ($host ? $host : null);
        $host = $hostname ? $hostname . '' . $port : '';
        $url_origin = $protocol . '://' . $host;
        return $url_origin;
    }

    /**
     * Adds child frames to $current_frame (searches $frames).
     * Function is recursive.
     * @return void
     */
    public static function setupFrameStack(&$current_frame, $frames)
    {
        if (!is_array($frames)) {
            return;
        }

        if (!is_array($current_frame)) {
            return;
        }

        // setup map w/ parent frame id as keys
        if (!isset(self::$traceMapByParentFrameId)) {
            self::$traceMapByParentFrameId = array();

            foreach ($frames as $frame_key => $frame) {


                if (array_key_exists($frame['call'], self::$defaultOptions['watches'])) {
                    $watchObject = self::$defaultOptions['watches'][$frame['call']];

                    // Add TRACKED_FUNC property
                    if (isset($watchObject['trackedFunction']) && $watchObject['trackedFunction'] == 'true') {
                        $frame['props']['TRACKED_FUNC'] = self::$defaultOptions['watches'][$frame['call']]['trackedFunctionName'];
                    }

                    // Set $customUrlTransactionName for Custom Reporting URL.
                    // Check if executed transaction's callback - get the custom transaction name for CUSTOM URL of the root frame specially if having nested  transaction functions call
                    if (self::$hasExecutedTransactionFunction) {
                        if (isset($watchObject['transactionName']) && $watchObject['transactionName'] != "" && $frame['props']['FRAME_ID'] == 1) {
                            self::$customUrlTransactionName = $watchObject['transactionName'];
                        }

                    } else {    // normal setting
                        if (isset($watchObject['transactionName']) && self::$customUrlTransactionName == null) {
                            self::$customUrlTransactionName = self::$defaultOptions['watches'][$frame['call']]['transactionName'];
                        }
                    }
                }

                if (isset($frame['props']['PARENT_FRAME_ID'])) {
                    if (!array_key_exists($frame['props']['PARENT_FRAME_ID'], self::$traceMapByParentFrameId)) {
                        self::$traceMapByParentFrameId[$frame['props']['PARENT_FRAME_ID']] = array();
                    }
                    array_push(self::$traceMapByParentFrameId[$frame['props']['PARENT_FRAME_ID']], $frame);
                }
            }
        }

        $current_frame['stacks'] = array();
        if (array_key_exists($current_frame['props']['FRAME_ID'], self::$traceMapByParentFrameId)) {
            $child_frames = self::$traceMapByParentFrameId[$current_frame['props']['FRAME_ID']];

            foreach ($child_frames as $child_frame_key => $child_frame) {
                self::setupFrameStack($child_frame, $frames);
                $current_frame['stacks'][] = $child_frame;
            }
        }
    }

    /**
     * Checks if file exceeds max size.
     * @param $filename
     * @return bool
     */
    public static function isFileOverSizeLimit($filename)
    {
        $max_file_size_bytes = 50 * 1024 * 1024;
        clearstatcache();
        return filesize($filename) > $max_file_size_bytes;
    }

    /**
     * Creates a new trace log file if current file is over limit.
     * Will also delete out older file in the process.
     * @param $filename_current_link_source - actual file path for current log file
     * @param $filename_current - symlink to current log file
     * @param $max_files - max amount of trace log files to keep
     */
    protected static function rollOverTraceLog($filename_current_link_source, $filename_current,  $max_files)
    {
        set_error_handler("Stackify\Profiler::stackifyErrorHandler");

        if (self::isFileOverSizeLimit($filename_current)) {

            Util\Sync::$lock = self::$logFilePath . DIRECTORY_SEPARATOR . '.lock';
            Util\Sync::call(function () use ($filename_current_link_source, $filename_current, $max_files) {
                // double check if file is over size
                if (self::isFileOverSizeLimit($filename_current_link_source)) {
                    // find the current # log file we are writing to. (name-#.log)
                    $index_start = strrpos($filename_current_link_source, "-");
                    $current_log_index = substr($filename_current_link_source, $index_start + 1, -4);

                    // delete oldest file
                    $max_file_index = $current_log_index - $max_files + 1;
                    if ($max_file_index > 0) {
                        $delete_filename = substr($filename_current_link_source, 0, $index_start) . "-" . $max_file_index . ".log";

                        if (file_exists($delete_filename)) {
                            if (Util::unLink($delete_filename) === false) {
                                self::log(1, "Profiler::rollOverTraceLog - Util\Sync::call - Unable to unlink delete_filename: ".$delete_filename);
                            }
                        }
                    }

                    // increment # to write to
                    $next_log_index = $current_log_index + 1;
                    $new_filename = substr($filename_current_link_source, 0, $index_start) . "-" . $next_log_index . ".log";

                    if (file_exists($new_filename)) {
                        if (Util::unLink($new_filename) === false) {
                            self::log(1, "Profiler::rollOverTraceLog - Util\Sync::call - Unable to unlink new_filename: ".$new_filename);
                        }
                    }

                    if (Util::isLink($filename_current)) {
                        if (Util::unLink($filename_current) === false) {
                            self::log(1, "Profiler::rollOverTraceLog - Util\Sync::call - Unable to unlink filename_current: ".$filename_current);
                        }
                    }

                    if (touch($new_filename) === false) {
                        self::log(1, "Profiler::rollOverTraceLog - Util\Sync::call - Unable to touch new_filename: ".$new_filename);
                    }

                    if (chmod($new_filename, 0777) === false) {
                        self::log(1, "Profiler::rollOverTraceLog - Util\Sync::call - Unable to change mod new_filename: ".$new_filename);
                    }

                    if (Util::makeSymlink($new_filename, $filename_current) === false) {
                        self::log(1, "Profiler::rollOverTraceLog - Util\Sync::call - Unable to link new_filename: ".$new_filename." to filename_current: ".$filename_current);
                    }

                    clearstatcache();
                }
            });
        }
        restore_error_handler();
    }

    /**
     * Removes any broken symlink log files
     */
    protected static function cleanupFiles($logFilePath)
    {
        set_error_handler("Stackify\Profiler::stackifyErrorHandler");

        Util\Sync::$lock = $logFilePath . DIRECTORY_SEPARATOR . '.lock';
        Util\Sync::call(function () use ($logFilePath) {
            foreach (scandir($logFilePath . DIRECTORY_SEPARATOR) as $entry) {
                $path = $logFilePath . DIRECTORY_SEPARATOR . $entry;
                if (Util::isLink($path) && file_exists($path) === false) {
                    if (Util::unLink($path) === false) {
                        self::log(1, "Profiler::cleanupFiles - Util\Sync::call - Unable to unlink path: ".$path);
                    }
                }
            }
        });

        restore_error_handler();
    }

    /**
     * Writes contents to trace log file.
     * @param $content
     */
    protected static function writeToTraceLog($content)
    {
        self::checkOrCreateLogFilePath();
        if (self::isRequiredDirOk()) {
            set_error_handler("Stackify\Profiler::stackifyErrorHandler");

            $logFilePath = self::$logFilePath;
            $threadId = "#0";
            $file_date_hash = md5(gmdate("YmdH"));
            $filename_base = gethostname() . "-" . $file_date_hash . $threadId;
            $filename_current = $logFilePath . DIRECTORY_SEPARATOR . $filename_base . ".log"; // symlink log file (for current file)

            // create a log file if one does not exist
            if (!file_exists($filename_current)) {

                Util\Sync::$lock = $logFilePath . DIRECTORY_SEPARATOR . '.lock';
                Util\Sync::call(function () use ($filename_current, $filename_base, $logFilePath) {
                    // double check
                    // ensure it still doesn't exist
                    if (file_exists($filename_current) === false) {
                        // delete all files w/ current hostname
                        $glob_result = glob($logFilePath . DIRECTORY_SEPARATOR . $filename_base . "*");
                        foreach ($glob_result as $f) {
                            if (file_exists($f) === true) {
                                if (Util::unLink($f) === false) {
                                    self::log(1, "Profiler::writeToTraceLog - Util\Sync::call - Unable to unlink file: " . $f);
                                }
                            }
                        }

                        if ($glob_result === false) {
                            self::log(1, "Profiler::writeToTraceLog - Util\Sync::call - Glob failed: " . $logFilePath . DIRECTORY_SEPARATOR . $filename_base . "*");
                        }

                        // create 1st file and symlink
                        if (touch($logFilePath . DIRECTORY_SEPARATOR . $filename_base . "-1.log") === false) {
                            self::log(1, "Profiler::writeToTraceLog - Util\Sync::call - Unable to touch file: " . $filename_base . "-1.log");
                        }

                        if (chmod($logFilePath . DIRECTORY_SEPARATOR . $filename_base . "-1.log", 0777) === false) {
                            self::log(1, "Profiler::writeToTraceLog - Util\Sync::call - Unable to change mod file: " . $filename_base . "-1.log");
                        }

                        if (Util::makeSymlink($logFilePath . DIRECTORY_SEPARATOR . $filename_base . "-1.log", $filename_current) === false) {
                            self::log(1, "Profiler::writeToTraceLog - Util\Sync::call - Unable to link file: " . $filename_base . "-1.log to file: " . $filename_current);
                        }
                        clearstatcache();
                    }
                });

            }

            $filename_current_link_source = readlink($filename_current);
            if ($filename_current_link_source === false) {
                self::log(1, "Profiler::writeToTraceLog - Unable to read symbolic link filename_current_link_source: " . $filename_current);
                return;
            }

            $file_put_contents = file_put_contents($filename_current_link_source, $content, FILE_APPEND | LOCK_EX);
            if ($file_put_contents === false) {
                self::log(1, "Profiler::writeToTraceLog - Unable to write contents to filename_current_link_source: " . $filename_current_link_source . " symlink: " . $filename_current);
            }

            // see if we have reached size limit
            // Only run this check every 25 requests
            if (rand(1, 25) == 25) {
                self::rollOverTraceLog($filename_current_link_source, $filename_current, self::$max_files);
            }

            if (rand(1, 100) == 100) {
                self::cleanupFiles(self::$logFilePath);
            }

            restore_error_handler();
        }
    }

    /**
     * Setup log directories
     *
     * @return void
     */
    public static function setupLogDirectories()
    {
        if (Util::isWindows() === false) {
            if (Util::isPrefixEnabled()) {
                self::$logFilePath =  "/usr/local/prefix/log";
                self::$errorLog = "/usr/local/prefix/debug/stackify-php-apm.log";
            } else {
                self::$logFilePath =  "/usr/local/stackify/stackify-php-apm/log";
                self::$errorLog = "/usr/local/stackify/stackify-php-apm/log/stackify-php-apm.log";
            }
        } else {
            $defaultDir = 'C:'.DIRECTORY_SEPARATOR.'ProgramData';
            $usersProfile = getenv('ALLUSERSPROFILE');
            $programData = getenv('PROGRAMDATA');

            if (!empty($usersProfiler)) {
                $defaultDir = $usersProfile;
            } else if (!empty($programData)) {
                $defaultDir = $programData;
            }

            $defaultDir = $defaultDir.DIRECTORY_SEPARATOR.'Stackify'.DIRECTORY_SEPARATOR.'PHP';
            self::$logFilePath =  $defaultDir.DIRECTORY_SEPARATOR.'Log';
            self::$errorLog = $defaultDir.DIRECTORY_SEPARATOR.'debug'.DIRECTORY_SEPARATOR.'stackify-php-apm.log';
        }
    }

    /**
     * Get log file path
     * 
     * @return string
     */
    public static function getLogFilePath()
    {
        return self::$logFilePath;
    }

    /**
     * auto start profiler for web and cli scenarios
     */
    public static function autoStart()
    {
        self::setupLogDirectories();

        if (self::isStarted() === false) {
            self::start();
        }
    }

    public static function isStarted()
    {
        return self::$mode !== self::MODE_DISABLED;
    }

    /**
     * Start profiling request.
     *
     * @param array $options
     */
    public static function start($options = array())
    {
        if (php_sapi_name() == 'cli') {
            $monitorCli = self::getEnvOrConfig('RETRACE_MONITOR_CLI', 'stackify.monitor_cli');
            ini_set('stackify.monitor_cli', $monitorCli);

            if ($monitorCli == false) {
                return;
            }

            $options = array(
                'service' => 'cli'
            );
            self::$isCli = true;
        }

        stackify_disable(); // this discards any data that was collected up to now and restarts.

        $rumEnabled = getenv('RETRACE_RUMV2_ENABLED');
        if (is_string($rumEnabled)) {
            ini_set('stackify.rum_enabled', $rumEnabled);
        } else {
            $rumEnabled = ini_get('stackify.rum_enabled');
        }

        $defaults = array(
            'collect' => self::MODE_FULL,
            'transport' => self::getEnvOrConfig('STACKIFY_TRANSPORT', 'stackify.transport'),
            'transport_http_endpoint' => self::getEnvOrConfig('STACKIFY_TRANSPORT_HTTP_ENDPOINT', 'stackify.transport_http_endpoint'),
            'application_name' => self::getEnvOrConfig('STACKIFY_APPLICATION_NAME', 'stackify.application_name'),
            'environment_name' => self::getEnvOrConfig('STACKIFY_ENVIRONMENT_NAME', 'stackify.environment_name'),
            'service' => self::getEnvOrConfig('STACKIFY_SERVICE', 'stackify.service'),
            'framework' => self::getEnvOrConfig('STACKIFY_FRAMEWORK', 'stackify.framework'),
            'rum_script_url' => self::getEnvOrConfig('RETRACE_RUM_SCRIPT_URL', 'stackify.rum_script_url'),
            'rum_key' => self::getEnvOrConfig('RETRACE_RUM_KEY', 'stackify.rum_key'),
            'rum_enabled' => $rumEnabled,
            'use_forwarded_host' => self::getEnvOrConfig('RETRACE_USE_FORWARDED_HOST', 'stackify.use_forwarded_host'),
            'monitor_cli' => self::$isCli
        );

        $options = array_merge($defaults, $options);

        self::init($options);
        self::enableProfiler();
    }
    
    /**
     * Get env or config value if not empty
     */
    private static function getEnvOrConfig($key, $config = null)
    {
        $envValue = getenv($key);
        if (!empty($envValue)) {
            return $envValue;
        }

        if (!$config) {
            return null;
        }

        return ini_get($config);
    }

    /**
     * Check required directories/folders/file permissions and availability, if something wrong don't proceed.
     *
     * @return boolean
     */
    private static function isRequiredDirOk()
    {
        // Check if log folder is present and writable
        if (self::$logFilePathWritable === null) {
            self::$logFilePathWritable = is_writable(self::$logFilePath);
            if (!self::$logFilePathWritable) {
                self::log(1, "Couldn't start PHP profiler: ". self::$logFilePath ." is not writable.");
            }
        }
        return self::$logFilePathWritable;
    }

    /**
     * Check or create log directory if does not exists
     *
     * @return void
     */
    private static function checkOrCreateLogFilePath()
    {
        if (self::$logFilePathExist !== null) {
            return;
        }

        if (@is_dir(self::$logFilePath) === false) {
            if (@mkdir(self::$logFilePath, 0777, true) == false) {
                self::log(1, "Couldn't start PHP profiler: Failed to create log folder - ". self::$logFilePath);
                self::$logFilePathExist = false;
                return;
            }

            @chmod(self::$logFilePath, 0777);
        }

        self::$logFilePathExist = true;
    }

    /**
     * Check or create error log directory if does not exists
     *
     * @return void
     */
    private static function checkOrCreateErrorLogPath()
    {
        if (self::$errorLogExist !== null) {
            return;
        }

        $errorLogDir = dirname(self::$errorLog);
        if (@is_dir($errorLogDir) === false) {
            if (@mkdir($errorLogDir, 0777, true) == false) {
                self::phpErrorLog(1, "Couldn't start PHP profiler: Failed to create error log folder - ". self::$logFilePath);
                self::$errorLogExist = false;
                return;
            }

            @chmod($errorLogDir, 0777);
        }

        self::$errorLogExist = true;
    }

    /**
     * Initialize profiler.
     *
     * @param $options
     */
    private static function init($options)
    {
        if ($options['framework']) {
            self::setupFramework($options['framework']);
        }

        self::$applicationName = $options['application_name'];
        self::$environmentName = $options['environment_name'];
        self::$transport = $options['transport'];
        self::$transportHttpEndpoint = $options['transport_http_endpoint'];
        self::$mode = self::MODE_FULL;
        self::$error = false;

        $rumScriptUrl = $options['rum_script_url'];
        if (preg_match('/^((((https?|ftps?|gopher|telnet|nntp):\/\/)|(mailto:|news:))(%[0-9A-Fa-f]{2}|[-()_.!~*\';\/\?:@&=+$,A-Za-z0-9])+)([).!\';\/\?:,][\[:blank:|:blank:\]])?$/', $rumScriptUrl)) {
            self::$rumScriptUrl = $rumScriptUrl;
        }

        $rumKey = $options['rum_key'];
        if (preg_match('/^[A-Za-z0-9_-]+$/', $rumKey)) {
            self::$rumKey = $rumKey;
        }

        self::$options = $options;
    }

    /**
     * Get options
     */
    public static function getOptions()
    {
        return self::$options;
    }

    /**
     * Adjusts profiler configuration based on passed in framework.
     * @param $framework
     */
    public static function setupFramework($framework)
    {
        self::$defaultOptions['framework'] = $framework;
    }

    /**
     * Enable the profiler.
     *
     * @return void
     */
    private static function enableProfiler()
    {
        self::$mode = self::MODE_FULL;
        self::$stackifyExtension = new \Stackify\Profiler\StackifyExtension();

        $setting = self::readCustomConfigFile('stackify.json','instrumentation');

        // default to empty array
        if (!is_array($setting)) {
            $setting = array();
        }

        // if aws lambda is detected
        $lambdaFunctioName = getenv('AWS_LAMBDA_FUNCTION_NAME');
        if (!empty($lambdaFunctioName)) {
            $lambdaHandler = getenv('_HANDLER');
            // track fast functions
            self::$defaultOptions['show_fast_functions'] = 1;
            // add handler function to custom instrumentation
            $lambda_call = array("class" => "", "method" => $lambdaHandler, "startTrace" => true);
            array_push($setting, $lambda_call);
        }

        self::map_stackify_json($setting);

        // Setting up transaction callback
        if( self::$isCli && !empty(self::$defaultOptions['transaction_functions']) ) {
            self::$defaultOptions = array_merge(self::$defaultOptions, array(
                'transaction_callback' => function() {
                    self::$hasExecutedTransactionFunction = true;

                    // Start linking frames here if there is atleast 1 function call sets as TRANSACTION FUNCTION
                    self::sendTrace();
                }
            ));
        }

        // check if track all is enabled
        if (self::isTrackAll()) {
            self::$defaultOptions['ignored_functions'] = array();
            self::$defaultOptions['show_fast_functions'] = 1;
        }

        stackify_enable(0, self::$defaultOptions);

        foreach (self::$defaultOptions['watches'] as $key => $val) {
            self::$stackifyExtension->watchSpan($key, $val['category']);
        }
    }

    /**
     * Get `stackify.error_reporting` filter
     *
     * @return int
     */
    private static function _getErrorReportingLevelFilter()
    {
        $stackifyErrorReporting = ini_get("stackify.error_reporting");
        $isNumeric = is_numeric($stackifyErrorReporting);

        if ($stackifyErrorReporting === "") {
            return null;
        }

        if ($isNumeric === false) {
            return null;
        }

        return intval($stackifyErrorReporting);
    }

    /**
     * Read custom JSON file then push to self::$defaultOptions['watches']
     * @param $filename
     * @param $propertyToRead
     * @return array
     */
    public static function readCustomConfigFile($filename, $propertyToRead)
    {
        $configFileLocations = array(
            ini_get('stackify.config_file'),
            static::getServerValue('DOCUMENT_ROOT').'/'.$filename,
            __DIR__.'/'.$filename
        );

        if (Util::isPrefixEnabled()) {
            $configFileLocations[] = '/usr/local/prefix/'.$filename;
        } else {
            $configFileLocations[] = '/usr/local/stackify/stackify-php-apm/'.$filename;
        }

        foreach ($configFileLocations as $fileDir) {
            if (Util::isFileNotEmpty($fileDir)) {
                try {
                    $content = file_get_contents($fileDir);
                }
                catch (\Exception $ignored) {
                }

                $json = json_decode($content, true);

                if ($json !== "") {
                    foreach ($json as $key => $val) {
                        if ($key == $propertyToRead) {
                            return $val;
                        }
                    }
                }
                break;
            }
        }
    }

    /**
     * Returns the mapped items from stackify.json. Expose for the user to review
     *
     * @return array
     */
    public static function get_watches()
    {
        return self::$defaultOptions['watches'];
    }

    /**
     * Iterates array from stackify.json then map all items
     *
     * @param $setting
     */
    public static function map_stackify_json($setting)
    {
        if (is_array($setting) || is_object($setting)) {
            foreach ($setting as $setting_key => $setting_props) {
                if (isset($setting_props['method']) && $setting_props['method'] != "" ) {
                    $category = "";  // Value might be 'View' | 'Event' but for now set this to empty
                    $sro = ($setting_props['class'] !== "") ? '::' : '';
                    $call = $setting_props['class'].$sro.$setting_props['method'];

                    $watches['call'] = $call;
                    $watches['category']  = $category;

                    // Checks if trackedFunction is set. Used for TRACKED FUNCTIONS
                    if (isset($setting_props['trackedFunction']) && $setting_props['trackedFunction'] != "" && is_bool($setting_props['trackedFunction'])) {   // if so, set the item
                        $watches['trackedFunction']  = ($setting_props['trackedFunction']) ? 'true' : 'false';
                        $str = (isset($setting_props['trackedFunctionName']) && $setting_props['trackedFunction'] == true && $setting_props['trackedFunction'] != "") ? $setting_props['trackedFunctionName']: "";
                        $watches['trackedFunctionName'] = self::create_template($call, $setting_props, $str);
                    } else {    // else don't set
                        unset($watches['trackedFunction']);
                        unset($watches['trackedFunctionName']);
                    }

                    // Checks if transactionName is set. Used for CUSTOM URL
                    if ( isset($setting_props['transactionName']) ) {
                        $str = (isset($setting_props['transactionName'])) ? $setting_props['transactionName'] : "";
                        $watches['transactionName']  = self::create_template($call, $setting_props, $str);
                    } else {
                        unset($watches['transactionName']);
                    }

                    // Checks if CLI, if so, set startTrace.
                    if ( self::$isCli && isset($setting_props['startTrace']) && is_bool($setting_props['startTrace']) ) {

                        $watches['startTrace']  = ($setting_props['startTrace']) ? 'true' : 'false';

                        // Map function call as transaction function. Avoid duplicate
                        if ( !in_array( $call, self::$defaultOptions['transaction_functions']) && $setting_props['startTrace'] ) {
                            array_push(self::$defaultOptions['transaction_functions'], $call);
                        }

                    } else {
                        unset($watches['startTrace']);
                    }

                    // Checks for duplicate function call.
                    if ( isset(self::$defaultOptions['watches'][$call]) ) {
                        $config_file = ini_get('stackify.config_file');
                        self::log(1, "You may have more than one setting of $call in $config_file, only one has been mapped. Use \Stackify\Profiler::get_watches() to review the successfully mapped custom instrumentation config.");
                    } else {
                        self::$defaultOptions['watches'][$call] = $watches;
                    }
                }
            }
        }
    }

    /**
     * Creates template by replacing {{ClassName}} and {{MethodName}} with class and method values.
     *
     * @param string $call
     * @param array $arr
     * @param string $string
     *
     * @return string
     */
    public static function create_template($call, $arr, $string)
    {
        if ($string != "") {
            $class = ($arr['class'] == "" || empty($arr['class'])) ? "php" : $arr['class'];
            return str_replace(array("{{ClassName}}", "{{MethodName}}"), array($class, $arr['method']), $string);
        } else {
            return $call;
        }
    }

    /**
     * Requires auto prepend file set
     *
     * @return bool
     */
    public static function requiresDelegateToOriginalPrependFile()
    {
        return ini_get('stackify.auto_prepend_library') &&
            stackify_prepend_overwritten() &&
            ini_get("auto_prepend_file") &&
            file_exists(stream_resolve_include_path(ini_get("auto_prepend_file")));
    }

    /**
     * Watch a function for calls and create timeline spans around it.
     *
     * @param string $function
     * @param string $category
     */
    public static function watch($function, $category = null)
    {
        if (!isset(self::$defaultOptions['watches'][$function])) { // not watched?
            self::$defaultOptions['watches'][$function] = $category;
            if (!self::$stackifyExtension->watchSpan($function, $category)) {
                self::log(1, "Unable to watch function: $function, something went wrong with the extension.");
            }
        }
    }

    /**
     * Log a message to the file defined in $errorLog
     *
     * @param int $level Logs message level. 1 = warning, 2 = info, 3 = debug
     * @param string $message
     *
     * @return void
     */
    public static function log($level, $message)
    {
        try {
            self::checkOrCreateErrorLogPath();

            if (self::$errorLogFile === null) {
                $errorLogFileWritable = @touch(self::$errorLog);

                if (self::$logFilePathWritable) {
                    if ($errorLogFileWritable === false) {
                        self::phpErrorLog(1, 'Stackify\Profiler::log() - Unable to create the log file: stackify-php-apm.log');
                    } else if (@chmod(self::$errorLog, 0777) === false) {
                        self::phpErrorLog(1, 'Stackify\Profiler::log() - Unable to permissions of the log file: stackify-php-apm.log');
                    }
                }

                self::$errorLogFile = $errorLogFileWritable;
            }

            if (self::$errorLogFile === true) {
                self::stackifyLog($level, $message);
            } else {
                self::phpErrorLog($level, $message);
            }
        } catch (\Exception $ex) {
            $errorLogFile = false;
            self::phpErrorLog(1, 'Unable able to log error to stackify-php-apm.log - '.$ex->getMessage());
        }
    }

    /**
     * Log message to stackify php apm log
     * @param int $level
     * @param string $message
     *
     * @return bool
     */
    private static function stackifyLog($level, $message)
    {
        $t = microtime(true);
        $ms = sprintf("%03d",($t - floor($t)) * 1000);
        $level = ($level === 3) ? "DEBUG" : (($level === 2) ? "INFO" : "WARN");
        return error_log(sprintf('%s %s [Stackify] - %s '.PHP_EOL, date('Y-m-d H:i:s,'.$ms), $level, $message), 3, self::$errorLog);
    }

    /**
     * Log message to php error log
     * @param int $level
     * @param string $message
     *
     * @return bool
     */
    private static function phpErrorLog($level, $message)
    {
        $t = microtime(true);
        $ms = sprintf("%03d",($t - floor($t)) * 1000);
        $level = ($level === 3) ? "DEBUG" : (($level === 2) ? "INFO" : "WARN");
        return error_log(sprintf('%s %s [Stackify] - %s', date('Y-m-d H:i:s,'.$ms), $level, $message), 0);
    }
    
    /**
     * Get insert rum script tag
     *
     * @return string|null
     */
    public static function insertRumScript()
    {
        try {
            if (!function_exists('stackify_transaction_id')) {
                return null;
            }

            $transactionId = stackify_transaction_id();
            if (empty($transactionId)) {
                return null;
            }

            if (empty(self::$rumScriptUrl) || empty(self::$rumKey)) {
                return null;
            }

            $rumSettings = array(
                'ID' => $transactionId
            );
    
            if (!empty(self::$applicationName)) {
                $rumSettings['Name'] = base64_encode(utf8_encode(self::$applicationName));
            }
    
            if (!empty(self::$applicationName)) {
                $rumSettings['Env'] = base64_encode(utf8_encode(self::$environmentName));
            }
    
            $reportingUrl = self::getReportingUrl();
            if (!empty($reportingUrl)) {
                $reportingUrl = \Stackify\Util::maskReportingUrl($reportingUrl);
                $requestMethod = static::getServerValue('REQUEST_METHOD');
                if (!self::$isCli && !empty($requestMethod)) {
                    $reportingUrl = "{$requestMethod}-{$reportingUrl}";
                }
                $rumSettings['Trans'] = base64_encode(utf8_encode($reportingUrl));
            }
    
            if (function_exists('stackify_transaction_mark_rum')) {
                stackify_transaction_mark_rum();
            }

            return '<script type="text/javascript">(window.StackifySettings || (window.StackifySettings = '.self::safe_json_encode($rumSettings).'))</script><script src="'.self::$rumScriptUrl.'" data-key="'.self::$rumKey.'" async></script>';
        } catch (\Exception $e) {
            self::log(3, "Stackify\Profiler::insertRumScript - Exception: ". $e->getMessage());
        }
    }

    /**
     * Get reporting url
     *
     * @return string
     */
    public static function getReportingUrl()
    {
        // allow transactionName to take precedence for reporting_url
        if (empty(self::$transactionName)) {
            $reporting_url = stackify_transaction_name();
        } else {
            $reporting_url = self::$transactionName;
        }

        // Custom URL from stackify.json setting
        if (empty($reporting_url)) {
            $customized_url = self::custom_url();
            $reporting_url = ($customized_url != null) ? $customized_url : $reporting_url;
        }

        if (empty($reporting_url)) {
            if (self::$isCli) {
                $reporting_url = basename(static::getServerValue('PHP_SELF'), '.php');
            } else {
                $reporting_url =  static::getServerValue('REQUEST_URI');
                if (!empty($reporting_url)) {
                    $reporting_url_parts = explode('?', $reporting_url);
                    if (!empty($reporting_url_parts)) {
                        $reporting_url = array_filter($reporting_url_parts)[0];
                    }
                }

                if (!empty($reporting_url)) {
                    $reporting_url = rtrim($reporting_url, '/');
                } 
                
                if (empty($reporting_url)) {
                    $reporting_url = '/'; // default
                }
            }
        }

        return $reporting_url;
    }

    /**
     * Get request headers
     *
     * @return array
     */
    public static function getRequestHeaders()
    {
        $requestHeaders = array();
        if (function_exists('getallheaders')) {
            foreach (getallheaders() as $name => $value) {
                $request_header = [
                    $name => $value
                ];
                array_push($requestHeaders, $request_header);
            }

            return $requestHeaders;
        }
        
        // Fallback
        // Referece: https://stackoverflow.com/a/41427998
        foreach ($_SERVER as $name => $value) {
            if (substr($name, 0, 5) == 'HTTP_') {
                $requestHeaders[str_replace(' ', '-', ucwords(strtolower(str_replace('_', ' ', substr($name, 5)))))] = $value;
            }
        }
        return $requestHeaders;
    }

    /**
     * Get server variable
     *
     * @param string $variable Key to search on server variable
     *
     * @return string|array
     */
    public static function getServerValue($variable)
    {
        if (!isset($GLOBALS['_SERVER'])) {
            return null;
        }

        if ($variable == null) {
            return $_SERVER;
        }

        return isset($_SERVER[$variable]) ? $_SERVER[$variable]: null;
    }

    /**
     * Get use forwarded host setting
     *
     * @return boolean
     */
    public static function useForwardedHost()
    {
        return static::$options['use_forwarded_host'] == '1';
    }
}

// auto-starts the profiler if that is configured
\Stackify\Profiler::autoStart();

/**
 * This is to allow the required file (from `auto_prepend_file`) to use the global
 * scope for the variables declared globally is inside the script.
 * 
 * Reference: https://www.php.net/manual/en/language.variables.scope.php
 */
if (\Stackify\Profiler::requiresDelegateToOriginalPrependFile()) {
    require_once ini_get("auto_prepend_file");
}